import unittest
from parameterized import parameterized
import pk_model as pk
import random

numReps = 1000

class DoseFnTests(unittest.TestCase):
    @parameterized.expand([(random.random()*100, random.random()*100+100, random.random()*10, random.random()*200) for j in range(numReps)])
    def test_PulseFn(self, start, stop, dose, check):
        doseFn = pk.blockPulse()
        doseFn.add_pulse(start, stop, dose)
        self.assertEqual(doseFn(check), dose if check >= start and check < stop else 0)



# from pandas.io.parsers import read_csv
# from matplotlib.lines import _LineStyle  # not sure if any of these needed
import numpy as np
from ..abstractPlotter import AbstractPlotter
from matplotlib import pyplot as plt
import pandas as pd


class PlotFromCSV(AbstractPlotter):
    '''
    Class which takes CSV-formatted data from a
    given file and returns a matplot,
    with the option to save figure
    '''

    def __init__(self, filename: str):
        '''
        Class is initialised by a .csv containing
        the datapoints, with headers giving the list
        of variable names that have been solved for
        in the methods section. This variable name
        list should lead with the time variable
        '''
        self.data_file = filename

    def plot(self, file_format):
        '''
        Function which takes a csv data file, reads it and separates the data
        into different variables including dose and time. Outputs a plot of
        each variable on the same plot and saves to a given file format
        '''
        full_data = pd.read_csv(self.data_file, delimiter=',')
        # Read_csv outputs a data frame first column is time, second column is
        # the dose at each timepoint.
        var_names = full_data.columns.tolist()
        var_number = len(var_names)
        full_data = pd.DataFrame(full_data).to_numpy()
        time = full_data[:, 0]
        dose = full_data[:, 1]
        data = full_data[:, 2:]
        fig, axes1 = plt.subplots()
        colours = plt.cm.viridis(np.linspace(0, .67, var_number))
        for i in range(var_number - 2):
            axes1.plot(time, data[:, i], color=colours[i], 
                       label=var_names[i + 2])
        axes1.plot([], [], color='r', linestyle='dashed', label='Dose')
        # recall the first two entries of var_names are time and dose,
        # so need shifted index

        axes1.set_ylabel('drug mass [ng]')
        axes1.legend()
        plt.xlabel('time [h]')
        axes2 = axes1.twinx()
        axes2.set_ylabel('dose [ng/h]')
        axes2.plot(time, dose, color='r', linestyle='dashed', label='Dose')
        axes1.legend()
        fig.tight_layout()
        file_root = self.data_file.split('.')[0]
        fig_file = str(file_root) + '.' + str(file_format)
        fig.savefig(fig_file)  # may need to change which directory it saves in
        plt.show()



"""Module containing the pharmokinetics (PK) subcutaneous (sub) parameters
class.

It contains a class with a getParam and params_set methods.
"""

from ..abstractParameters import AbstractParameters


class SubCutNCompParameters(AbstractParameters):
    """Class in which the parameters for the PK sub model are stored as a
    dictionary.

    It inherits from AbstractParameters.
    """

    def __init__(self, nCompartments, *args, **kwargs):

        expect_params = \
            ['V_c', 'CL', 'k_a', 'q_c0', 'q_e0'] + \
            ['Q_p{}'.format(i) for i in range(1, nCompartments+1)] +\
            ['V_p{}'.format(i) for i in range(1, nCompartments+1)] +\
            ['q_p{}_0'.format(i) for i in range(1, nCompartments+1)]
        for param in expect_params:
            assert param in kwargs, 'Missing parameter "{}"'.format(param)
        self.params = {key: kwargs[key] for key in kwargs}

    def getParam(self, key: str) -> float:
        """Returns the parameter associated to a certain key in the dictionary.

        :param key: name associated to the parameter to retrieve
        :return: parameter associated to the key
        :rtyoe: float
        """
        assert key in self.params
        return self.params[key]

    def getParameterNames(self) -> list:
        """Returns the complete list of parameters, in the form of a dictionary.
        :return: list of parameters and their keys
        :rtype: dictionary
        """
        return self.params.keys()

    def __str__(self) -> str:
        """Convert to string detailing parameters stored in the dictionary.
        : return : Description of parameters
        : rtype : str
        """
        return str(self.params)



"""Module containing the pharmokinetics abstract data collector class.
"""


from abc import ABC, abstractmethod
import numpy as np


class AbstractDataCollector(ABC):
    """Class in which the Data Collector is abstractly defined.

    The class will receive data from the model (at each timestep) and will save
    the data to a numpy array, with an option to write to a csv file.
    """

    @abstractmethod
    def begin(self, names: list, number_timesteps: int):
        """Initialises a numpy array of a given size, ready to receive data.

        :param names: list of the names of the parameters,
                      used as column headers
        :param number_timesteps: number of time steps in the model resolution,
                                 used to define the columns length
        """
        pass

    @abstractmethod
    def report(self, data: np.ndarray) -> np.array:
        """Report data as a column vector into an array at each timestep.

        :param data: numpy array containing the data of the model resolution
        :return: numpy array containing the model solution
        """
        pass

    @abstractmethod
    def __getitem__(self, i: int) -> np.ndarray:
        """Return data as a column vector at a time point requested. Asserts
        timepoint is within the 'past' of the model.
        """
        pass

    @abstractmethod
    def writeToFile(self, filename: str) -> None:
        """Opens a given filename and writes data in csv format.

        :param filename: name of the file in which the data will be stored
        """
        pass





import unittest
from unittest.mock import Mock, patch
from parameterized import parameterized
import pk_model as pk
import random
import numpy as np
import string
import builtins

def generateFnHelperNumpyArray(n: int):
    def inner():
        return np.random.random((n, 1))
    return inner

subjects = [pk.NumpyDataCollector]
generateFnHelpers = [generateFnHelperNumpyArray]
compareFns = [lambda a,b: a.all()==b.all()]

class DataCollectorTests(unittest.TestCase):
    @parameterized.expand([(s,) for s in subjects])
    def test_Construct(self, subject):
        testSubject = subject()
        testSubject.begin(self.helperCols(2), 5)
    
    @parameterized.expand([(s,t,c,i) for s, t in zip(subjects, generateFnHelpers)
        for c, i in zip([5, 10]+[random.randint(10, 100) for j in range(20)], [10, 50]+[random.randint(50, 100) for j in range(20)])])
    def test_reportTest(self, subject, generateFnHelper, numCols: int, numIters: int):
        testSubject = subject()
        testSubject.begin(self.helperCols(numCols), numIters)
        self.mockReport(testSubject, generateFnHelper(numCols), numIters)

    @parameterized.expand([(s,t,cmp,c,i) for s, t, cmp in zip(subjects, generateFnHelpers, compareFns)
        for c, i in zip([5, 10]+[random.randint(10, 100) for j in range(20)], [10, 50]+[random.randint(50, 100) for j in range(20)])])
    def test_reportRetrieveTest(self, subject, generateFnHelper, compareFn, numCols: int, numIters: int):
        testSubject = subject()
        testSubject.begin(self.helperCols(numCols), numIters)
        data = self.mockReport(testSubject, generateFnHelper(numCols), numIters)
        self.mockRetrieveAndVerify(testSubject, data, compareFn)

    @parameterized.expand([(s,t) for s, t in zip(subjects, generateFnHelpers)])
    def test_WriteFile(self, subject, generateFnHelper):
        testSubject = subject()
        nCols, nIters = random.randint(2,10), random.randint(10, 100)
        testSubject.begin(self.helperCols(nCols), nIters)
        self.mockReport(testSubject, generateFnHelper(nCols), nIters)
        filename = ''.join(random.choice(string.ascii_letters) for i in range(15))
        with patch('builtins.open', unittest.mock.mock_open()) as mock_open:
            testSubject.writeToFile(filename)
        mock_open.assert_called_once_with(filename, 'w')
        handle = mock_open()
        self.assertEqual(handle.write.call_count, nIters + 1, msg='Incorrect number of write calls')
    
    @parameterized.expand([(s,t) for s, t in zip(subjects, generateFnHelpers)])
    def test_FailRetrieve(self, subject, generateFnHelper):
        testSubject = subject()
        nCols, nIters = 2, 5
        testSubject.begin(self.helperCols(nCols), nIters)
        self.helperTestFailRetrieve(testSubject, generateFnHelper(nCols))
    
    def helperTestFailRetrieve(self, testSubject, generateFn):
        self.assertRaises(AssertionError, testSubject.__getitem__, -1)
        self.assertRaises(AssertionError, testSubject.__getitem__, 0)
        self.mockReport(testSubject, generateFn, 3)
        testSubject.__getitem__(2)
        testSubject.__getitem__(0)
        self.assertRaises(AssertionError, testSubject.__getitem__, 3)
        self.assertRaises(AssertionError, testSubject.__getitem__, -2)
    
    def helperCols(self, numCols: int) -> list:
        return [''.join(random.choice(string.ascii_letters) for i in range(10)) for j in range(numCols)]
    
    def mockReport(self, testSubject, generateFn, numIterations):
        data = []
        for i in range(numIterations):
            data.append(generateFn())
            testSubject.report(data[i])
        return data
    
    def mockRetrieveAndVerify(self, testSubject, data, compareFn):
        for i in range(len(data)):
            retrieved = testSubject[i]
            self.assertTrue(compareFn(data[i], retrieved),
                'Expected {}, but {} was returned.'.format(data[i], retrieved))



from .models.iv_model_scipy import IvModelScipy
from .models.sub_model_scipy import SubModelScipy
from .models.iv_model_ncompt_scipy import NComptIvModelScipy
from .models.sub_model_ncompt_scipy import NComptSubModelScipy


class ModelFactory():
    def getIvModelScipy():
        return IvModelScipy
    
    def getSubModelScipy():
        return SubModelScipy
    
    def getNCompIVModel():
        return NComptIvModelScipy
    
    def getNCompSubCutModel():
        return NComptSubModelScipy


import unittest
import pk_model as pk

class TestAbstractClasses(unittest.TestCase):
    def test_AbstractDataCollector(self):
        pk.AbstractDataCollector.begin(None, 0, 0)
        pk.AbstractDataCollector.report(None, 0)
        pk.AbstractDataCollector.__getitem__(None, 0)
        pk.AbstractDataCollector.writeToFile(None, '')

    def test_AbstractParameters(self):
        pk.AbstractParameters.getParam(None, '')
        pk.AbstractParameters.getParameterNames(None)
        pk.AbstractParameters.__str__(None)
    
    def test_AbstractModel(self):
        pk.AbstractModel.solve(None)

    def test_AbstractPlotter(self):
        pk.AbstractPlotter.plot(None, '')



"""pkmodel is a Pharmokinetic modelling library.

It contains functionality for creating, solving, and visualising the solution
of Parmokinetic (PK) models

"""
# Import version info
from .version_info import VERSION_INT, VERSION  # noqa



# Import main abstract classes
from .abstractModel import AbstractModel    # noqa
from .abstractParameters import AbstractParameters    # noqa
from .abstractDataCollector import AbstractDataCollector     # noqa
from .abstractPlotter import AbstractPlotter

from .api import solve_iv_toFile, solve_model_from_config, solve_subcut_toFile,\
    create_expDecay_dosing, create_periodic_dosing, create_singlePulse_dosing,\
    solve_model_from_config, process_config,\
    plot_single_file, plot_varying_parameter


# Import Parameters
from .parameters.parameters_iv import IV_Parameters
from .parameters.parameters_sub import Subcut_Parameters
from .parameters.iv_nCompartment_parameters import IVNCompartmentParameters
from .parameters.sub_nCompartment_parameters import SubCutNCompParameters

# Import Models
from .models.iv_model_scipy import IvModelScipy
from .models.sub_model_scipy import SubModelScipy
from .models.iv_model_bckEuler import IVModelBckEuler
from .models.sub_model_bckEuler import SubModelBckEuler
from .models.sub_model_ncompt_scipy import NComptSubModelScipy
from .models.iv_model_ncompt_scipy import NComptIvModelScipy

# Import Data Collectors
from .dataCollectors.dataCollector_numpy import NumpyDataCollector

# Import Dose Functions
from .Block_pulse_dose import blockPulse

# Import Plotters
from .plotters.plotFromCSV import PlotFromCSV
from .plotters.plotFromConfig import PlotFromConfig




#
# Version information for pkmodel.
#
# See: https://packaging.python.org/guides/single-sourcing-package-version/
#
# Version as a tuple (major, minor, revision)
#  - Changes to major are rare
#  - Changes to minor indicate new features, possible slight backwards
#    incompatibility
#  - Changes to revision indicate bugfixes, tiny new features
VERSION_INT = 1, 0, 0

# String version of the version number
VERSION = '.'.join([str(x) for x in VERSION_INT])




class blockPulse():

    """ A block dosing function that can add multiple block pulses of a dose,
        each dose pulse has a beginning time, end time and dose ammount
    """
    def __init__(self):
        self.pulses = {}
   
    def add_pulse(self, start_time, stop_time, dose) -> None:
        """ Method to add individual pulses
            Parameters:
            start_time: time at beginning of dosing pulse
            stop_time: time at end of dosing pulse
            dose: concentration of dose
        """
        self.pulses[start_time, stop_time] = dose
        return

    def __call__(self, x : float) -> float:
        """ Calls dosing function to be used in solution of PK models
        """
        current = 0
        for key in self.pulses:
            pulseAmmount = self.pulses[key]
            startTime = key[0]
            stopTime=key[1]
            if x >= startTime and x < stopTime:
                current += pulseAmmount
        return current


from .parameters.parameters_iv import IV_Parameters
from .parameters.parameters_sub import Subcut_Parameters
from .parameters.iv_nCompartment_parameters import IVNCompartmentParameters
from .parameters.sub_nCompartment_parameters import SubCutNCompParameters


class ParametersFactory():
    def getIVParameters():
        return IV_Parameters

    def getSubcutParameters():
        return Subcut_Parameters

    def getNCompIVParameters():
        return IVNCompartmentParameters
    
    def getNCompSubCutParameters():
        return SubCutNCompParameters


import unittest
import pk_model as pk
import random

class TestParameters(unittest.TestCase):
    """
    Tests the 'AbstractParameters' classes
    """
    
    def helperTestConstruct(self, testClass: pk.AbstractParameters, paramNames: set) -> None:
        testParams = {key: random.random() for key in paramNames}
        testSubject = testClass(**testParams)

    def helperTestRetrieve(self, testClass: pk.AbstractParameters, paramNames: set) -> None:
        testParams = {key: random.random() for key in paramNames}
        testSubject = testClass(**testParams)
        for key in testParams:
            self.assertEqual(testParams[key], testSubject.getParam(key))

    def helperTestStr(self, testClass: pk.AbstractParameters, paramNames: set) -> None:
        testParams = {key: random.random() for key in paramNames}
        testSubject = testClass(**testParams)
        self.assertEqual(str(testParams), str(testSubject))

    def helperTestGetKeys(self, testClass: pk.AbstractParameters, paramNames: set) -> None:
        testParams = {key: random.random() for key in paramNames}
        testSubject = testClass(**testParams)
        self.assertEqual(testParams.keys(), testSubject.getParameterNames())

    def test_IVParameters(self):
        self.helperTestConstruct(pk.IV_Parameters, {'Q_p', 'V_c', 'V_p', 'CL', 'q_c0', 'q_p0'})
        self.helperTestRetrieve(pk.IV_Parameters, {'Q_p', 'V_c', 'V_p', 'CL', 'q_c0', 'q_p0'})
        self.assertRaises(AssertionError, self.helperTestConstruct, pk.IV_Parameters, {'Q_p', 'V_c'})
        self.helperTestStr(pk.IV_Parameters, {'Q_p', 'V_c', 'V_p', 'CL', 'q_c0', 'q_p0'})
        self.helperTestGetKeys(pk.IV_Parameters, {'Q_p', 'V_c', 'V_p', 'CL', 'q_c0', 'q_p0'})

    def test_SubCutParameters(self):
        self.helperTestConstruct(pk.Subcut_Parameters, {'Q_p', 'V_c', 'V_p', 'CL', 'k_a', 'q_c0', 'q_p0', 'q_e0'})
        self.helperTestRetrieve(pk.Subcut_Parameters, {'Q_p', 'V_c', 'V_p', 'CL', 'k_a', 'q_c0', 'q_p0', 'q_e0'})
        self.assertRaises(AssertionError, self.helperTestConstruct, pk.Subcut_Parameters, {'Q_p', 'V_c'})
        self.helperTestStr(pk.Subcut_Parameters, {'Q_p', 'V_c', 'V_p', 'CL', 'k_a', 'q_c0', 'q_p0', 'q_e0'})
        self.helperTestGetKeys(pk.Subcut_Parameters, {'Q_p', 'V_c', 'V_p', 'CL', 'k_a', 'q_c0', 'q_p0', 'q_e0'})
    
    def test_NIVParameters(self):
        testSubject = pk.IVNCompartmentParameters(0, Q_p=0, V_c=0, CL=0, q_c0=1)
        str(testSubject)
        testSubject.getParameterNames()

    def test_NSubCutParameters(self):
        testSubject = pk.SubCutNCompParameters(0, Q_p=0, V_c=0, CL=0, q_c0=1, k_a=0, q_e0=1)
        str(testSubject)
        testSubject.getParameterNames()


#
# Abstract Plotting class
#

from abc import ABC, abstractmethod


class AbstractPlotter(ABC):
    """An abstract class which
    can be adapted to various
    plotting methods

    """

    @abstractmethod
    def plot(self, filename: str) -> None:
        """
        Plot data in file
        """
        pass





"""Module containing the pharmokinetics (PK) subcutaneous (sub) model class.

It contains a class with a solve method, in which the equations
of the model are defined and solved.
"""

import typing
import numpy as np
import scipy.integrate

from ..abstractModel import AbstractModel
from ..abstractParameters import AbstractParameters
from ..abstractDataCollector import AbstractDataCollector


class NComptSubModelScipy(AbstractModel):
    """Class in which the PK sub model is defined and solved.

    It inherits from AbstractModel.
    It contains the function solve and the subfunction pk_sub_model.
    """

    def __init__(
        self,
        parameters: AbstractParameters,
        solution: AbstractDataCollector,
        dosefunction: typing.Callable[[float], float],
        timespan: float,
        nsteps: int,
        ncompartments: int,
    ):
        self.parameters = parameters
        self.solution = solution
        self.dosefunction = dosefunction
        self.timespan = timespan
        self.nsteps = nsteps
        self.ncompartments = ncompartments

    def solve(self):
        """Solves the three compartments PK sub model and outputs to solution.
        It gets the parameters using the parameter class method.
        The solver used is scipy.

        It writes line by line the solution of the ODEs using the solution
        class method.
        Solution format: [time, dose, q_e, q_c, q_p].
        """

        # Definition of the parameters
        V_c = self.parameters.getParam("V_c")
        CL = self.parameters.getParam("CL")
        k_a = self.parameters.getParam("k_a")

        initial_conditions = [0 for i in range(self.ncompartments + 2)]
        initial_conditions[0] = self.parameters.getParam("q_e0")
        initial_conditions[1] = self.parameters.getParam("q_c0")
        for i in range(1, self.ncompartments + 1):
            initial_conditions[i + 1] = self.parameters.getParam("q_p{}_0".format(i))

        t_eval = np.linspace(0, self.timespan, self.nsteps)

        # Definition of the model ODEs
        def pk_sub_model(t, q):
            """Defines the differential equations for the PK sub model.

            Parameters:
            :param t: time (h)
            :param y: list of the state variables of the ODEs system, in the
                      form [q_e, q_c, q_p]
            :param Q_p: transition rate between central and peripheral
                         compartments (mL/h)
            :param V_c: volume of central compartment (mL)
            :param V_p: volume of peripheral compartment (mL)
            :param CL: clearance/elimination rate from the central
                       compartment (mL/h)
            :param k_a: absorption rate in the subcutaneous model (/h)

            The parameters (except for t and y) are extracted from the
            Parameter class, using getParam method.

            Returns list containing the differential equations, in the form:
            [dqe_dt, dqc_dt, dqp_dt]
            """
            result = [0 for i in range(2 + self.ncompartments)]
            result[0] = self.dosefunction(t) - q[0] * k_a
            result[1] = q[0] * k_a - q[1] * CL / V_c
            for i in range(1, 1 + self.ncompartments):
                Q_p = self.parameters.getParam("Q_p{}".format(i))
                V_p = self.parameters.getParam("V_p{}".format(i))
                result[i + 1] = Q_p * (q[1] / V_c - q[i + 1] / V_p)
                result[1] -= result[i + 1]
            return result

        # Solving the model
        sol = scipy.integrate.solve_ivp(
            fun=pk_sub_model,
            t_span=[t_eval[0], t_eval[-1]],
            y0=initial_conditions,
            t_eval=t_eval,
        )

        # Feeding the solution line by line to solution class
        t = sol.t
        y = sol.y
        N = t.shape[0]
        columnNames = ["t", "dose", "q_e", "q_c"] + [
            "q_p{}".format(i) for i in range(1, self.ncompartments + 1)
        ]
        self.solution.begin(columnNames, N)
        for i in range(N):
            arr = np.zeros((len(columnNames), 1))
            arr[0] = t[i]
            arr[1] = self.dosefunction(t[i])
            arr[2:, 0] = y[:, i]
            self.solution.report(arr)



import unittest
from unittest.mock import Mock, MagicMock, call, patch, ANY
import pk_model as pk
import random
import string
from parameterized.parameterized import parameterized
import numpy as np
import pandas as pd
import json
from .config import *
import builtins

numReps = 25
doseFns = [lambda t: 0, lambda t: 1, lambda t: t, lambda t: t*t]

class ApiTest(unittest.TestCase):
    @parameterized.expand([() for i in range(numReps)])
    def test_solve_iv_toFile(self):
        outfilename = ''.join(random.choice(string.ascii_letters) for j in range(15))
        Q_p, V_c, V_p, CL, q_c0, q_p0 = random.random(),random.random(),random.random(),random.random(),random.random(),random.random()
        doseFn=lambda x: 0
        tSpan, numIters= random.random()*1000, random.randint(100, 1000)
        
        with patch('pk_model.IV_Parameters.__init__', return_value=None) as mockMakeParams,\
                patch('pk_model.NumpyDataCollector.__init__', return_value=None) as mockMakeCollector,\
                patch('pk_model.IvModelScipy.__init__', return_value=None) as mockMakeModel,\
                patch('pk_model.IvModelScipy.solve') as mockSolve,\
                patch('pk_model.NumpyDataCollector.writeToFile') as mockWriteFile:
            pk.solve_iv_toFile(outfilename=outfilename, Q_p=Q_p, V_c=V_c, V_p=V_p, CL=CL, q_c0=q_c0, q_p0=q_p0,
                doseFn=doseFn, tSpan=tSpan, numIters=numIters)
        mockMakeParams.assert_called_once_with(Q_p=Q_p, V_c=V_c, V_p=V_p, CL=CL, q_c0=q_c0, q_p0=q_p0)
        mockMakeCollector.assert_called_once_with()
        mockMakeModel.assert_called_once_with(ANY, ANY, doseFn, tSpan, numIters)
        mockSolve.assert_called_once_with()
        mockWriteFile.assert_called_once_with(outfilename)

    def test_solve_subcut_toFile(self):
        outfilename = ''.join(random.choice(string.ascii_letters) for j in range(15))
        Q_p, V_c, V_p, CL, k_a, q_e0, q_c0, q_p0 = random.random(), random.random(), random.random(),random.random(),random.random(),random.random(),random.random(),random.random()
        doseFn=lambda x: 0
        tSpan, numIters= random.random()*1000, random.randint(100, 1000)
        
        with patch('pk_model.Subcut_Parameters.__init__', return_value=None) as mockMakeParams,\
                patch('pk_model.NumpyDataCollector.__init__', return_value=None) as mockMakeCollector,\
                patch('pk_model.SubModelScipy.__init__', return_value=None) as mockMakeModel,\
                patch('pk_model.SubModelScipy.solve') as mockSolve,\
                patch('pk_model.NumpyDataCollector.writeToFile') as mockWriteFile:
            pk.solve_subcut_toFile(outfilename=outfilename, Q_p=Q_p, V_c=V_c, V_p=V_p, CL=CL, k_a=k_a, q_e0=q_e0, q_c0=q_c0, q_p0=q_p0,
                doseFn=doseFn, tSpan=tSpan, numIters=numIters)
        mockMakeParams.assert_called_once_with(Q_p=Q_p, V_c=V_c, V_p=V_p, CL=CL, k_a=k_a, q_e0=q_e0, q_c0=q_c0, q_p0=q_p0)
        mockMakeCollector.assert_called_once_with()
        mockMakeModel.assert_called_once_with(ANY, ANY, doseFn, tSpan, numIters)
        mockSolve.assert_called_once_with()
        mockWriteFile.assert_called_once_with(outfilename)

    @parameterized.expand([() for i in range(numReps)])   
    def test_create_expDecay_dosing(self):
        A, k = random.random()*10, random.random()
        testFn = pk.create_expDecay_dosing(A, k)
        t = random.random()*100
        self.assertEqual(testFn(t), A*np.exp(-np.abs(k)*t))
    
    @parameterized.expand([() for i in range(numReps)])
    def test_create_periodic_dosing(self):
        tHigh, tLow = random.random()*2, random.random()*2
        highVal, lowVal = random.random(), random.random()
        testFn = pk.create_periodic_dosing(tHigh, tLow, highVal, lowVal)
        t = random.random()*100
        self.assertEqual(testFn(t), highVal if t%(tHigh+tLow) <= tHigh else lowVal)

    def test_create_singlePulse_dosing(self):
        tStart, tStop = random.random()*2, random.random()+3
        val = random.random()
        testFn = pk.create_singlePulse_dosing(tStart, tStop, val)
        t = random.random()*5
        self.assertEqual(testFn(t), val if t >= tStart and t < tStop else 0)

    @parameterized.expand([(c,d) for c in modelConfigsSuccess for d in doseFns])
    def test_solveModelFromConfig(self, cfg, doseFn):
        with patch('builtins.open', unittest.mock.mock_open()) as mock_open:
            pk.solve_model_from_config(cfg, doseFn)

    @parameterized.expand([(c, d) for c in modelConfigsFail for d in doseFns])
    def test_solveModelFromConfigFail(self, cfg, doseFn):
        with patch('builtins.open', unittest.mock.mock_open()) as mock_open:
            self.assertRaises(AssertionError, pk.solve_model_from_config, cfg, doseFn)

    @parameterized.expand([({"modelConfig": {}},d, []) for d in doseFns])
    def test_processConfig(self, cfg, doseFn, outfiles):
        with patch('builtins.open', unittest.mock.mock_open()) as mock_open, \
            patch('json.load', return_value = cfg) as cfgPatch, \
            patch('pk_model.api.solve_model_from_config', return_value = outfiles), \
            patch('matplotlib.pyplot'):
            pk.process_config("config", doseFn)
    
    @parameterized.expand([(c,d, []) for d in doseFns for c in fullConfigSuccess])
    def test_processConfigWithPlot(self, cfg, doseFn, outfiles):
        with patch('builtins.open', unittest.mock.mock_open()) as mock_open, \
            patch('json.load', return_value = cfg) as cfgPatch, \
            patch('pk_model.PlotFromCSV.plot'),\
            patch('pandas.read_csv', return_value = pd.DataFrame(data={'t':[1,2],'d':[2,3],'a':[3,4]})):
            pk.process_config("config", doseFn)

    @parameterized.expand([({"aaa": {}},d, []) for d in doseFns])
    def test_processConfigFails(self, cfg, doseFn, outfiles):
        with patch('builtins.open', unittest.mock.mock_open()) as mock_open, \
                patch('json.load', return_value = cfg) as cfgPatch, \
                patch('pk_model.api.solve_model_from_config', return_value = outfiles):
            self.assertRaises(AssertionError,pk.process_config,"config", doseFn)




#
# Abstract Model class
#

from abc import ABC, abstractmethod


class AbstractModel(ABC):
    """A Pharmokinetic (PK) model

    Parameters
    ----------

    value: numeric, optional
        an example paramter

    """

    @abstractmethod
    def solve(self):
        """
        Solve using parameters and write output to solution
        """
        pass




import numpy as np

from ..abstractPlotter import AbstractPlotter
from matplotlib import pyplot as plt
import pandas as pd
import os


class PlotFromConfig(AbstractPlotter):
    '''
    Class which takes CSV-formatted data from a
    given file and returns a matplot,
    with the option to save figure
    '''

    def __init__(self, config_dir: dir, filenames: str):
        '''
        Class is initialised by a list of .csv files containing
        the datapoints, and a directory of the configurable
        parameters given in the setup file.
        '''
        self.data_files = filenames
        self.config = config_dir

    def plot(self):
        '''
        Function which takes a csv data files, reads it and separates the data
        into different variables including dose and time. Outputs plots
        comparing given variables wth the option to include the Dose function
        and change line colours. It then saves to a given file format.
        '''
        # list the variables output that wants to be plotted
        vars_to_plot = self.config["plotVars"]

        for var in vars_to_plot:
            #Initialise an empty graph, and empty data matrix
            i = 0
            fig, axes1 = plt.subplots()
            for file in self.data_files:
                # Each run outputs in a separate .csv file so
                # loop through the information in each file
                full_run_data = pd.read_csv(file, delimiter=',')
                var_names = full_run_data.columns.tolist()
                # Variable names and data position may change
                # depending on the model used.

                full_data = pd.DataFrame(full_run_data).to_numpy()
                time = full_data[:, 0]
                index = var_names.index(var)
                data = full_data[:, index]
                if not self.config["ParameterLabels"]:
                    Label = "run" + '=' + str(i + 1)
                else:
                    Label = self.config["ParameterLabels"][i]
                if len(self.config["ParameterLineColour"]) != len(self.data_files):
                    colours = plt.cm.viridis(np.linspace(0, .67,
                                             len(self.data_files)))
                    colour = colours[i]
                else:
                    colour = self.config["ParameterLineColour"][i]

                axes1.plot(time, data, label=Label,
                           color=colour)
                i += 1
            if self.config["DoseLine"]:
                if len(self.config["DoseLineColour"]) == 0:
                    Dosecolor = 'r'
                elif len(self.config["DoseLineColour"]) == 1:                   
                    Dosecolor = self.config["DoseLineColour"][0]
                else:
                    Dosecolor = self.config["DoseLineColour"][i]
                # if loop plots the Dose function on a second set of axes
                axes1.plot([], [], color=Dosecolor, linestyle='dashed',
                           label='Dose')
                axes2 = axes1.twinx()
                axes2.set_ylabel('dose [ng/h]')
                axes2.plot(time, full_data[:, 1], color=Dosecolor,
                           linestyle='dashed', label='Dose')

            # Formatting for the full figure
            axes1.set_ylabel('drug mass [ng]')
            axes1.legend()
            plt.xlabel('time [h]')
            axes1.legend()
            fig.tight_layout()
            file_dir = os.path.dirname(self.data_files[0])
            file_root = str(var)
            fig_file = str(file_root) + '.' + str(self.config["ImageFileFormat"])
            fig_file = os.path.join(file_dir, fig_file)
            # Figure is saved using the plotted variable name,
            # and saved in the given format.
            fig.savefig(fig_file)
            plt.show()
            fig.clf()



import unittest
from unittest.mock import Mock, MagicMock, call
import numpy as np
import pk_model as pk
from parameterized.parameterized import param, parameterized
import random

numReps = 10
ivSubjects = [pk.IvModelScipy, pk.IVModelBckEuler]
subSubjects = [pk.SubModelScipy, pk.SubModelBckEuler]

class ModelBasicTest(unittest.TestCase):
    
    @parameterized.expand([(s, random.randint(1, 10), random.randint(100, 1000)) for j in range(numReps) for s in ivSubjects+subSubjects])
    def test_construct(self, subject: pk.AbstractModel, endTime: float, numIters: int):
        subject(self.mockParameters, self.mockDataCollector(1), lambda x: 1, endTime, numIters)

    @parameterized.expand([(s, random.randint(1, 10), random.randint(100, 1000)) for j in range(numReps) for s in ivSubjects])
    def test_solveModelIV(self, subject: pk.AbstractModel, endTime: float, numIters: int):
        params = self.mockParameters()
        collector = self.mockDataCollector(4)
        testSubject = subject(params, collector, lambda x: 1, endTime, numIters)
        testSubject.solve()
        for c in [call('Q_p'), call('V_c'), call('V_p'), call('CL'), call('q_c0'), call('q_p0')]:
            self.assertIn(c, params.getParam.call_args_list)
        self.assertEqual(collector.report.call_count, numIters, 'Wrong number of points reported')

    @parameterized.expand([(s, random.randint(1, 10), random.randint(100, 1000)) for j in range(numReps) for s in subSubjects])
    def test_solveModelSci(self, subject: pk.AbstractModel, endTime: float, numIters: int):
        params = self.mockParameters()
        collector = self.mockDataCollector(5)
        testSubject = subject(params, collector, lambda x: 1, endTime, numIters)
        testSubject.solve()
        for c in [call('Q_p'), call('V_c'), call('V_p'), call('CL'), call('k_a'), call('q_e0'), call('q_c0'), call('q_p0')]:
            self.assertIn(c, params.getParam.call_args_list)
        self.assertEqual(collector.report.call_count, numIters, 'Wrong number of points reported')
    
    @parameterized.expand([(n, t, i) for n, t, i in zip(
        [random.randint(1, 5) for j in range(numReps)],
        [random.random()*10 for j in range(numReps)],
        [random.randint(10, 100) for j in range(numReps)]
    )])
    def test_solveModelNIV(self, nCompartments: int, endTime: float, numIters: int):
        params = self.mockParameters()
        collector = self.mockDataCollector(3+nCompartments)
        testSubject = pk.NComptIvModelScipy(params, collector, lambda x: 1, endTime, numIters, nCompartments)
        testSubject.solve()
        for c in [call('V_c'), call('CL'), call('q_c0')] + \
                [call('V_p{}'.format(i)) for i in range(1, nCompartments + 1)] + \
                [call('Q_p{}'.format(i)) for i in range(1, nCompartments+1)] + \
                [call('q_p{}_0'.format(i)) for i in range(1, nCompartments + 1)]:
            self.assertIn(c, params.getParam.call_args_list)
        self.assertEqual(collector.report.call_count, numIters, 'Wrong number of points reported')

    @parameterized.expand([(n, t, i) for n, t, i in zip(
        [random.randint(1, 5) for j in range(numReps)],
        [random.random()*10 for j in range(numReps)],
        [random.randint(10, 100) for j in range(numReps)]
    )])
    def test_solveModelNSubCut(self, nCompartments: int, endTime: float, numIters: int):
        params = self.mockParameters()
        collector = self.mockDataCollector(4+nCompartments)
        testSubject = pk.NComptSubModelScipy(params, collector, lambda x: 1, endTime, numIters, nCompartments)
        testSubject.solve()
        for c in [call('V_c'), call('CL'), call('k_a'), call('q_e0'), call('q_c0')] + \
                [call('V_p{}'.format(i)) for i in range(1, nCompartments + 1)] + \
                [call('Q_p{}'.format(i)) for i in range(1, nCompartments+1)] + \
                [call('q_p{}_0'.format(i)) for i in range(1, nCompartments + 1)]:
            self.assertIn(c, params.getParam.call_args_list)
        self.assertEqual(collector.report.call_count, numIters, 'Wrong number of points reported')


    def mockParameters(self):
        mock = Mock()
        mock.getParam.return_value = 1
        return mock

    def mockDataCollector(self, numCols):
        mock = MagicMock()
        mock.__getitem__.return_value = np.zeros((numCols, 1))
        return mock
        


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pk_model',
 'pk_model.dataCollectors',
 'pk_model.models',
 'pk_model.parameters',
 'pk_model.plotters',
 'pk_model.tests']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.3,<4.0.0',
 'numpy>=1.20.0,<1.21.0',
 'pandas>=1.3.4,<2.0.0',
 'scipy>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'pk-model',
    'version': '0.1.6',
    'description': 'Basic Pharmokinetic Model',
    'long_description': None,
    'author': 'Kit Gallagher',
    'author_email': 'gallagherkit@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



# Model solving the IV dosing set of equations, using backward euler
import typing
import dataclasses
import numpy as np

from ..abstractModel import AbstractModel
from ..abstractParameters import AbstractParameters
from ..abstractDataCollector import AbstractDataCollector


@dataclasses.dataclass
class SubModelBckEuler(AbstractModel):
    parameters: AbstractParameters
    dataCollector: AbstractDataCollector
    doseFn: typing.Callable[[float], float]
    tStop: float
    numIters: int

    def __post_init__(self):
        self.dt = self.tStop / self.numIters

    def eqMatrix(self, t):
        """Generate the Evolution Matrix M and Bias vector b, where M*q_next = q_prev + b
        :param t: float, time
        :returns : np.ndarray matrix M, np.ndarray column vector b
        """
        Q_p = self.parameters.getParam("Q_p")
        V_c = self.parameters.getParam("V_c")
        V_p = self.parameters.getParam("V_p")
        CL = self.parameters.getParam("CL")
        k_a = self.parameters.getParam("k_a")

        M = np.zeros((3, 3))
        M[0, 0] = 1 + self.dt * k_a
        M[1, 0] = -self.dt * k_a
        M[1, 1] = 1 + self.dt * CL / V_c + self.dt * Q_p / V_c
        M[1, 2] = self.dt * Q_p / V_p
        M[2, 1] = -self.dt * Q_p / V_c
        M[2, 2] = 1 + self.dt * Q_p / V_p

        b = np.zeros((3, 1))
        b[0, 0] = self.doseFn(t + self.dt) * self.dt
        return M, b

    def solve(self):
        self.dataCollector.begin(
            names=["t", "dose", "q_e", "q_c", "q_p"], number_timesteps=self.numIters
        )
        self.dataCollector.report(
            np.array(
                [
                    0,
                    self.doseFn(0),
                    self.parameters.getParam("q_e0"),
                    self.parameters.getParam("q_c0"),
                    self.parameters.getParam("q_p0"),
                ],
                ndmin=2,
            ).transpose()
        )

        for i in range(1, self.numIters):
            prev = self.dataCollector[i - 1]
            t = prev[0, 0] + self.dt
            dose = self.doseFn(t)

            M, b = self.eqMatrix(t)
            next = np.matmul(np.linalg.inv(M), prev[2:, [0]] + b)

            self.dataCollector.report(np.vstack((t, dose, next)))



import unittest
from unittest.mock import Mock, MagicMock, call
import numpy as np
import pk_model as pk
from parameterized.parameterized import param, parameterized
import random
import typing

reps = 10
ivSubjects = [pk.IvModelScipy, pk.IVModelBckEuler]
subSubjects = [pk.SubModelScipy, pk.SubModelBckEuler]

def generateIVParams(reps: int):
    return [pk.IV_Parameters(Q_p=1,V_c=1,V_p=1,CL=1,q_c0=random.random(),q_p0=random.random())for j in range(reps)]

def generateSubParams(reps: int):
    return [pk.Subcut_Parameters(Q_p=1,V_c=1,V_p=1,CL=1,k_a=1,q_e0=0,q_c0=0,q_p0=0)for j in range(reps)]



def generateDoseFns(reps: int):
    def getDoseFn():
        a = random.random()
        b = random.random()
        def inner(t: float) -> float:
            return a*t + b*t*t
        return inner
    return [getDoseFn() for j in range(reps)]

class ModelCompareTest(unittest.TestCase):

    @parameterized.expand([(s1, s2, p, d, t, n) for s1 in ivSubjects for s2 in ivSubjects if s1!=s2 for p, d, t, n in zip(generateIVParams(reps), generateDoseFns(reps), [random.random()/1000 for j in range(reps)], [random.randint(1000, 2000) for j in range(reps)])])
    def test_ivCompare(self, subject1: pk.AbstractModel, subject2: pk.AbstractModel, params: pk.AbstractParameters, doseFn: typing.Callable[[float], float], tSpan: float, numIters: int):
        sol1, sol2 = pk.NumpyDataCollector(), pk.NumpyDataCollector()
        testSubject1 = subject1(params, sol1, doseFn, tSpan, numIters)
        testSubject2 = subject2(params, sol2, doseFn, tSpan, numIters)
        testSubject1.solve()
        testSubject2.solve()
        self.helperCompare(sol1, sol2, numIters)

    @parameterized.expand([(s1, s2, p, d, t, n) for s1 in subSubjects for s2 in subSubjects if s1!=s2 for p, d, t, n in zip(generateSubParams(reps), generateDoseFns(reps), [random.random()/1000 for j in range(reps)], [random.randint(1000, 2000) for j in range(reps)])])
    def test_subCompare(self, subject1: pk.AbstractModel, subject2: pk.AbstractModel, params: pk.AbstractParameters, doseFn: typing.Callable[[float], float], tSpan: float, numIters: int):
        sol1, sol2 = pk.NumpyDataCollector(), pk.NumpyDataCollector()
        testSubject1 = subject1(params, sol1, doseFn, tSpan, numIters)
        testSubject2 = subject2(params, sol2, doseFn, tSpan, numIters)
        testSubject1.solve()
        testSubject2.solve()
        self.helperCompare(sol1, sol2, numIters)

    def helperCompare(self, sol1, sol2, numIters):
        point1, point2 = sol1[numIters-1], sol2[numIters-1]
        for i in range(point1.shape[0]):
            self.assertAlmostEqual(point1[i,0], point2[i,0], 1, "Models don't agree")


    




#
# Test module for pkmodel.
#
# To run all tests, use ``python -m unittest discover``.
#
# To run a particular test, use e.g.
#  ``python -m unittest pkmodel.tests.test_model``.
#
# For this project, we chose to place the tests inside the code directory. See
# https://stackoverflow.com/questions/61151/where-do-the-python-unit-tests-go
# for an extended discussion.
#




from .plotters.plotFromConfig import PlotFromConfig
from .plotters.plotFromCSV import PlotFromCSV

import typing
from .model_factory import ModelFactory
from .dataCollector_factory import DataCollectorFactory
from .parameters_factory import ParametersFactory
from .plotter_factory import PlotterFactory

import numpy as np
from .Block_pulse_dose import blockPulse
import json


def solve_iv_toFile(
    outfilename,
    Q_p=1,
    V_c=1,
    V_p=1,
    CL=1,
    q_c0=0,
    q_p0=0,
    doseFn=lambda x: 0,
    tSpan=1,
    numIters=1000,
):
    """Solve IV Model with given parameters and write solution at each
    iteration to outfile.

    :param V_c: the volume of the central compartment (mL)
    :param V_p: the volume of the peripheral compartment (mL)
    :param Q_p: the transition rate between central compartment and peripheral compartment
    :param CL: the clearance/elimination rate from the central compartment (mL/h)
    :param q_c0: the initial drug quantity in the central compartment (ng)
    :param q_p0: the initial drug quantity in the central compartment (ng)


    Usage:
    >>> import pkmodel as pk
    >>> pk.solve_iv_toFile("outData.csv",
                            Q_p=1, V_c=1,
                            doseFn=lambda t: 1,
                            tSpan=100,
                            numIters=1000)
    >>> import pandas as pd
    >>> pd.read_csv("outData.csv")
                      t  dose       q_c       q_p
    0      0.0000   1.0  0.000000  0.000000
    1      0.1001   1.0  0.090864  0.004540
    2      0.2002   1.0  0.166024  0.016509
    3      0.3003   1.0  0.228892  0.033884
    4      0.4004   1.0  0.282140  0.055095
    ..        ...   ...       ...       ...
    995   99.5996   1.0  0.999833  1.000103
    996   99.6997   1.0  0.999866  1.000083
    997   99.7998   1.0  0.999881  1.000073
    998   99.8999   1.0  0.999890  1.000068
    999  100.0000   1.0  0.999906  1.000058

    [1000 rows x 4 columns]

    """
    params = ParametersFactory.getIVParameters()(
        Q_p=Q_p, V_c=V_c, V_p=V_p, CL=CL, q_c0=q_c0, q_p0=q_p0
    )
    soln = DataCollectorFactory.getNumpyDataCollector()()
    model = ModelFactory.getIvModelScipy()(params, soln, doseFn, tSpan,
                                           numIters)
    model.solve()
    soln.writeToFile(outfilename)


def solve_subcut_toFile(
    outfilename,
    Q_p=1,
    V_c=1,
    V_p=1,
    CL=1,
    k_a=1,
    q_e0=0,
    q_c0=0,
    q_p0=0,
    doseFn=lambda x: 0,
    tSpan=1,
    numIters=1000,
):
    """Solve sub Model with given parameters and write solution at each
    iteration to outfile

    :param V_c: the volume of the central compartment (mL)
    :param V_p: the volume of the peripheral compartment (mL)
    :param Q_p: the transition rate between central compartment and peripheral compartment
    :param CL: the clearance/elimination rate from the central compartment (mL/h)
    :param q_c0: the initial drug quantity in the central compartment (ng)
    :param q_p0: the initial drug quantity in the central compartment (ng)
    :param k_a: the “absorption” rate from the entrance compartment for the subcutaneous dosing (/h)
    :param q_e0: the initial drug quantity in the entrance compartment (ng)

    Usage:

    >>> import pkmodel as pk
    >>> pk.solve_subcut_toFile("outData.csv", Q_p=1, V_c=1, doseFn=lambda t: 1,
    tSpan=100, numIters=1000)
    >>> import pandas as pd
    >>> pd.read_csv("outData.csv")
                            t  dose       q_e       q_c       q_p
    0      0.0000   1.0  0.000000  0.000000  0.000000
    1      0.1001   1.0  0.095253  0.004540  0.000151
    2      0.2002   1.0  0.181433  0.016511  0.001099
    3      0.3003   1.0  0.259404  0.033883  0.003372
    4      0.4004   1.0  0.329948  0.055102  0.007281
    ..        ...   ...       ...       ...       ...
    995   99.5996   1.0  1.000000  1.000024  0.999985
    996   99.6997   1.0  1.000000  1.000276  0.999829
    997   99.7998   1.0  1.000000  1.000539  0.999667
    998   99.8999   1.0  1.000000  1.000717  0.999557
    999  100.0000   1.0  1.000000  1.000687  0.999575

    [1000 rows x 5 columns]


    """
    params = ParametersFactory.getSubcutParameters()(
        Q_p=Q_p, V_c=V_c, V_p=V_p, CL=CL, q_c0=q_c0, q_p0=q_p0, k_a=k_a,
        q_e0=q_e0
    )
    
    soln = DataCollectorFactory.getNumpyDataCollector()()
    model = ModelFactory.getSubModelScipy()(params, soln, doseFn, tSpan,
                                            numIters)
    model.solve()
    soln.writeToFile(outfilename)
    


def create_expDecay_dosing(A: float, k: float):
    """Create an Exponentially decaying dosing profile (A*exp(-k*t))

    :param A float: Dose at t=0
    :param k float: Dosing decay rate
    :returns: Dosing Profile
    :rtype: Callable[[float], float]
    """

    def inner(t: float) -> float:
        return A * np.exp(-np.abs(k) * t)

    return inner

def create_singlePulse_dosing(tStart: float, tStop: float, dose: float) -> typing.Callable[[float], float]:
    """Create a single pulse dosing profile
    # ------------------------------------------------------------
    # Create a dosing profile which gives a
    # Remains high for timeHigh and low for lowHigh each period
    # ------------------------------------------------------------
    : param: tStart float: Time dosing starts
    : param: tStop float: Time dosing stops
    : param: dose: dose level
    : returns: Dosing Profile
    : rtype: Callable[[float], float]
    """
    ret = blockPulse()
    ret.add_pulse(tStart, tStop, dose)
    return ret

def create_periodic_dosing(timeHigh, timeLow, highVal, lowVal=0):
    """Create a Periodic dosing profile

    Create a dosing profile which oscillates between high and low values.
    Remains high for timeHigh and low for lowHigh each period

    :param timeHigh: Time dosing remains at highVal (float)
    :param timeLow: Time dosing remains at lowVal (float)
    :param highVal: dosing ammount during high levels (float)
    :param lowVal: dosing level during low levels (float)
    :returns: Dosing Profile
    :rtype: Callable[[float], float]
    """

    def inner(t: float) -> float:
        phase = t%(timeHigh + timeLow)
        return highVal if phase <= timeHigh else lowVal

    return inner


def plot_single_file(filename: str, format = 'png'):
    """Creates a single plot comparing the drug concentrations in different
    compartments. The dose rate is given on a second axis for comparison.

    :param filename: a .csv file (str)
    :param format: required image file format (str)
    :returns: graph saved to a png
    """

    figure = PlotterFactory.getPlotFromCSV()(filename)
    figure.plot(format)


def plot_varying_parameter(config_dir: dir, filenames: list):
    """ Takes a configuration directory containing various initial setups
    and outputs a graphs comparing different parameters across those files.

    :param filenames: list of strings of filenames, all in the same directory
    :param config_dir: directory containing configurable information, including
        the variable being compared across files.
    :returns: multiple graphs for each given variable. All saved to png files
    """

    figure = PlotFromConfig(config_dir, filenames)
    figure.plot()

def create_singlePulse_dosing(tStart: float, tStop: float, dose: float) -> typing.Callable[[float], float]:
    """Create a single pulse dosing function.

    :param tStart: time of start of the pulse (float)
    :param tStop: time of stop of the pulse (float)
    :param dose: dose quantity concentration (float)
    """
    ret = blockPulse()
    ret.add_pulse(tStart, tStop, dose)
    return ret


def solve_model_from_config(cfg: dict, doseFn: typing.Callable[[float], float]) -> typing.List[str]:
    """ Solve a model defined by a ModelConfig dictionary.
    Solves a specified N-compartment model for a timespan and number of
    iterations with protocol(s) specified in config.
    Returns list of output filenames which contain the data from the protocols.

    :param cfg: Model Configuration in a dictionary
                (i.e. read in from a json file)
    :param doseFn: Dosing function for models
    :returns: List of output file strings.
    :rtype: typing.List[str]

    Usage:
    
    >>> import pkmodel as pk
    >>> import json
    >>> cfg = json.load(open('modelConfig.json', 'r'))
    >>> pk.solve_model_from_config(cfg, lambda t: 1)
        ['protocol1Output.csv', 'protocol2Output.csv']
    """
    # Check model is specified in config
    assert "model" in cfg, "Config is missing \"model\", to specify model type."
    model = cfg["model"].lower().strip()
    # Check specified model is recognised
    assert model in ["iv", "subcut"], "Unrecognised model type."
    paramClass = ParametersFactory.getNCompIVParameters() if model == "iv"\
        else ParametersFactory.getNCompSubCutParameters()
    modelClass = ModelFactory.getNCompIVModel() if model == "iv" \
        else ModelFactory.getNCompSubCutModel()

    # Check required configs are set
    assert "tspan" in cfg, "Model config is missing 'tspan'"
    assert "numIterations" in cfg, "Model config is missing 'numIterations'"
    assert "numCompartments" in cfg, "Model config is missing 'numCompartments'"
    tspan, numIterations = cfg["tspan"], cfg["numIterations"]
    numCompartments = cfg["numCompartments"]
    
    # Check one of protocol or protocols is defined
    if "protocol" in cfg:
        protocols = [cfg["protocol"]]
    else:
        assert "protocols" in cfg, "ModelConfig must contain a 'protocol' or 'protocols' section."
        protocols = cfg["protocols"]

    outfiles = []
    # Run the protocols
    for protocol in protocols: # Might want to consider making this multithreaded
        params = paramClass(numCompartments, **protocol)
        collector = DataCollectorFactory.getNumpyDataCollector()()
        model = modelClass(params, collector, doseFn, tspan, numIterations, numCompartments)
        model.solve()
        outfile = protocol["outfile"] if "outfile" in protocol \
            else "tmp_{}".format(str(protocol))
        collector.writeToFile(outfile)
        outfiles.append(outfile)

    return outfiles

def process_config(configfile: str, doseFn: typing.Callable[[float], float]):
    """ Run a process defined by a config file.
    The config file defines a model which is to be run, with customisable
    number of protocols.
    The Dose Function is supplied as a separate argument

    Config file must contain a "modelConfig" section, which defines the model.
    This section must contain tspan, numIterations, numCompartments and either
    of protocol or protocols.
    Config file can also contain a section to define how the solutions should
    be plotted.
    
    :param configfile: filename of config json file (str)
    :param doseFn: Dose function
    
    Usage:

    >>> import pkmodel as pk
    >>> pk.process_config('config.json', pk.create_singlePulse_dosing(0, 0.1, 1))

    """
    cfg = json.load(open(configfile, 'r'))
    assert "modelConfig" in cfg, "Config file is missing modelConfig section."
    outfiles = solve_model_from_config(cfg["modelConfig"], doseFn)

    if "plotConfig" in cfg:
        if len(outfiles) == 1:
            plot_single_file(outfiles[0])
        elif len(outfiles) > 1:
            plot_varying_parameter(cfg["plotConfig"], outfiles)




"""Module containing the pharmokinetics data collector class.

It contains a class with begin, report and writetofile methods.
"""

import numpy as np
from ..abstractDataCollector import AbstractDataCollector


class NumpyDataCollector(AbstractDataCollector):
    """Class receives data from model (at each timestep) and saves the data to
    a numpy array, with an option to write to a csv file.
    """

    def begin(self, names: list, number_timesteps: int):
        """Initialises a numpy array of a given size, ready to receive data.

        :param names: list of the names of the parameters,
                      used as column headers
        :param number_timesteps: number of time steps in the model resolution,
                                 used to define the columns length
        """
        self.column_length = number_timesteps
        self.column_headers = names
        self.row_width = len(names)
        self.__content = np.zeros((number_timesteps, self.row_width))
        self.__index = 0

    def report(self, data: np.ndarray) -> np.array:
        """Report data as a column vector into an array at each timestep.

        :param data: numpy array containing the data of the model resolution
        :return: numpy array containing the model solution
        """
        assert data.shape == (self.row_width, 1), 'Invalid Data Shape'
        assert self.__index < self.column_length, \
               'Too many datapoints reported'
        self.__content[self.__index, :] = np.transpose(data)
        self.__index += 1

    def __getitem__(self, time_point: int) -> np.ndarray:
        """Return data as a column vector at a time point requested. Asserts
        timepoint is within the 'past' of the model.

        :param time_point: specified time at which we want the data
        :return: data as a column for the specified time step
        :rtype: numpy array column
        """
        assert (
            time_point < self.__index
            and time_point >= 0
            and time_point <= self.column_length
        )
        return np.transpose(self.__content[[time_point], :])

    def writeToFile(self, filename: str):
        """Opens a given filename and writes data in csv format.

        :param filename: name of the file in which the data will be stored
        """
        with open(filename, "w") as f:
            f.write(",".join(self.column_headers) + '\n')
            for i in range(self.column_length):
                f.write(",".join([str(x) for x in self.__content[i, :]])
                        + '\n')



"""Module containing the pharmokinetics abstract parameters class.
"""

from abc import ABC, abstractmethod
import typing

class AbstractParameters(ABC):
    """Class in which the methods of the parameters class are
    abstractly defined for the pharmakinetic protocol model.

    It inherits from AbstractParameters.
    """

    @abstractmethod
    def getParam(self, key: str) -> float:
        """Retrieve Parameter
        : param key : parameter name to retrieve
        : return : parameter value
        : rtype : float
        """
        pass

    @abstractmethod
    def getParameterNames(self) -> typing.Set[str]:
        """Retrieve Parameter Names
        : return : Set of parameter names
        : rtype : set[str]
        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Convert to string detailing parameters stored
        : return : Description of parameters
        : rtype : str
        """
        pass
